@@ -9,7 +9,11 @@ from sip.SipParser import parseBytes, buildMessage
 import common.util as util
 import common.client as client
 import sip.SipFlows as flow
-from copy import copy
+from threading import Lock
+
+
+def dialog_hash(dialog):
+    return "|".join([dialog["Call-ID"], dialog["from_tag"], dialog["to_tag"]])
 
 
 class SipEndpoint(object):
@@ -42,18 +46,20 @@ class SipEndpoint(object):
             "to_tag": None
             # "epid": None,
         }
-
+        self.tags = {}
         self.current_transaction = {"via_branch": None,
                                     "cseq": "0",
                                     "method": None
                                     }
 
         self.waitForMessage = self.wait_for_message  # compatibility alias
+        self.reply_to = self.send_in_ctx_of  # method alias
         self.secondary_lines = []
         self.message_buffer = []
         self.registered = False
         self.re_register_timer = None
         self.busy = False
+        self.lock = Lock()
 
     def update_parameters(self, params, force=False):
         """
@@ -94,11 +100,15 @@ class SipEndpoint(object):
         :param in_dialog: the complete dialog
         :return: None
         """
-        for dialog in self.dialogs:
-            if dialog["Call-ID"] == in_dialog["Call-ID"] \
-                    and dialog["from_tag"] == in_dialog["from_tag"] \
-                    and not dialog["to_tag"]:
-                dialog["to_tag"] = in_dialog["to_tag"]
+        with self.lock:
+            for dialog in self.dialogs:
+                if dialog["Call-ID"] == in_dialog["Call-ID"] \
+                        and dialog["from_tag"] == in_dialog["from_tag"] \
+                        and not dialog["to_tag"]:
+                    dhash = dialog_hash(dialog)
+                    dialog["to_tag"] = in_dialog["to_tag"]
+                    dhash_complete = dialog_hash(dialog)
+                    self.tags[dhash_complete] = self.tags[dhash]
 
     def get_complete_dialog(self, in_dialog):
         """
@@ -109,12 +119,13 @@ class SipEndpoint(object):
         """
         if in_dialog["to_tag"]:
             return in_dialog
-        for dialog in self.dialogs:
-            if dialog["Call-ID"] == in_dialog["Call-ID"] \
-                    and dialog["from_tag"] == in_dialog["from_tag"] \
-                    and dialog["to_tag"]:
-                return dialog
-        return in_dialog
+        with self.lock:
+            for dialog in self.dialogs:
+                if dialog["Call-ID"] == in_dialog["Call-ID"] \
+                        and dialog["from_tag"] == in_dialog["from_tag"] \
+                        and dialog["to_tag"]:
+                    return dialog
+            return in_dialog
 
     def set_address(self, address):
         """
@@ -156,6 +167,8 @@ class SipEndpoint(object):
         if dialog not in self.dialogs:
             self.dialogs.append(dialog)
             self.requests.append([])
+            # If we don't know this dialog it means we didn't started so it must be an incoming Request
+            self.tags[dialog_hash(dialog)] = "to_tag"
         dialog = self.get_complete_dialog(dialog)
         self.current_dialog = dialog
         self.parameters["callId"] = self.current_dialog["Call-ID"]
@@ -163,6 +176,18 @@ class SipEndpoint(object):
         self.parameters["toTag"] = self.current_dialog["to_tag"]
         return dialog
 
+    def switch_tags(self, dialog=None):
+        if not dialog:
+            dialog = self.current_dialog
+        dhash = dialog_hash(dialog)
+        dialog["from_tag"], dialog["to_tag"] = dialog["to_tag"], dialog["from_tag"]
+        dhash_switched = dialog_hash(dialog)
+        self.set_dialog(dialog)
+        if self.tags[dhash] == "to_tag":
+            self.tags[dhash_switched] = "from_tag"
+        else:
+            self.tags[dhash_switched] = "to_tag"
+
     def set_transaction(self, transaction):
         """ Change current transaction to the one provided """
         if not isinstance(transaction, dict):
@@ -182,6 +207,7 @@ class SipEndpoint(object):
             # "epid": lambda x=6: "SC" + util.randStr(x),
         }
         self.current_dialog = dialog
+        self.tags[dialog_hash(dialog)] = "from_tag"
         self.dialogs.append(dialog)
         self.requests.append([])
         return dialog
@@ -191,25 +217,27 @@ class SipEndpoint(object):
         # If we have received no messages yet return None
         if self.current_dialog == {"Call-ID": None, "from_tag": None, "to_tag": None}:
             return None
-        # First check for complete dialogs
-        for message in self.last_messages_per_dialog:
-            if message.get_dialog() == dialog:
-                return message
-        # Second check for incomplete dialogs
-        for message in self.last_messages_per_dialog:
-            d = message.get_dialog()
-            if d["Call-ID"] == dialog["Call-ID"] and d["from_tag"] == dialog["from_tag"]:
-                return message
+        with self.lock:
+            # First check for complete dialogs
+            for message in self.last_messages_per_dialog:
+                if message.get_dialog() == dialog:
+                    return message
+            # Second check for incomplete dialogs
+            for message in self.last_messages_per_dialog:
+                d = message.get_dialog()
+                if d["Call-ID"] == dialog["Call-ID"] and d["from_tag"] == dialog["from_tag"]:
+                    return message
         raise Exception("No message found in dialog {}. Other dialogs active: ".format(dialog, self.dialogs))
 
     def save_message(self, message):
         """ Search for previously received message in the same dialog.
             If found, replace with given message, otherwise append message to message list """
-        for i in range(len(self.last_messages_per_dialog)):
-            if message.get_dialog() == self.last_messages_per_dialog[i].get_dialog():
-                self.last_messages_per_dialog[i] = message
-                return
-        self.last_messages_per_dialog.append(message)
+        with self.lock:
+            for i in range(len(self.last_messages_per_dialog)):
+                if message.get_dialog() == self.last_messages_per_dialog[i].get_dialog():
+                    self.last_messages_per_dialog[i] = message
+                    return
+            self.last_messages_per_dialog.append(message)
 
     def start_new_transaction(self, method, dialog=None):
         """ Refresh the via branch and CSeq header """
@@ -219,27 +247,24 @@ class SipEndpoint(object):
             last_message_in_dialog = self.get_last_message_in(dialog)
         except:
             last_message_in_dialog = None
+        branch = util.randomBranch()
         if method in ("ACK", "CANCEL"):
             # Not really a new transaction
             # find transaction from last message in dialog
+            # which should not be none because what are we sending ACK or CANCEL to?
             transaction = last_message_in_dialog.get_transaction()
             cseq = transaction["cseq"]
-            branch = transaction["via_branch"]
         elif last_message_in_dialog:
             transaction = last_message_in_dialog.get_transaction()
             cseq = str(int(transaction["cseq"]) + 1)
-            branch = util.randomBranch()
         else:
-            cseq = str(len(self.requests[self.dialogs.index(dialog)]))
-            branch = util.randomBranch()
-            if method != "REGISTER":
-                # Ignore REGISTER otherwise unregister breaks
-                # TODO: check if reRegistrations will work
-                self.requests[self.dialogs.index(dialog)].append(method)
-        transaction = {}
-        transaction["via_branch"] = branch
-        transaction["cseq"] = cseq  # str(int(self.current_transaction["cseq"]) + 1)
-        transaction["method"] = method
+            with self.lock:
+                cseq = str(len(self.requests[self.dialogs.index(dialog)]))
+                if method != "REGISTER":
+                    # Ignore REGISTER otherwise unregister breaks
+                    # TODO: check if reRegistrations will work
+                    self.requests[self.dialogs.index(dialog)].append(method)
+        transaction = {"via_branch": branch, "cseq": cseq, "method": method}
         self.current_transaction = transaction
         return transaction
 
@@ -301,38 +326,42 @@ class SipEndpoint(object):
 
     def reply(self, message_string, dialog=None):
         """ Send a response to a previously received message """
-        if dialog:
-            dialog = self.set_dialog(dialog)
-        else:
-            dialog = self.current_dialog
-
         if "callId" not in self.parameters or not self.parameters['callId']:
             raise Exception("Cannot reply when we are not in a dialog")
 
         m = buildMessage(message_string, self.parameters)
 
+        if dialog:
+            dialog = self.set_dialog(dialog)
+        else:
+            dialog = self.current_dialog
+
         try:
             previous_message = self.get_last_message_in(dialog)
             m.make_response_to(previous_message)
             self.update_to_tag(m.get_dialog())
-        except:
+        except:  # caused by get_last_message_in(dialog) if no message has been exchanged in this dialog yet
             # New dialog, same call-id, eg NOTIFY after Keyset SUBSCRIBE.
-            # Must be a SIP Response
+            # Must be a SIP Request
+            previous_message = None
             assert m.type == "Request", \
                 "Attempted to send a {} response in a new dialog".format(m.get_status_or_method())
-            m.set_dialog_from(dialog)
 
         if m.type == "Request":
             # This is a new request in the same dialog, so fix the CSeq
             # m.increase_cseq()
             self.start_new_transaction(m.method)
+            # If B side sends the new request we must switch from and to tags
+            if previous_message is not None and \
+                    m.method not in ("ACK", "CANCEL") and \
+                    self.tags[dialog_hash(dialog)] == "to_tag":
+                self.switch_tags(dialog)
+            m.set_dialog_from(dialog)
+        else:
+            self.free_resources(m)
         self.save_message(m)
 
         m.set_transaction_from(self.current_transaction)
-
-        # TODO: fix CSeq according to RFC3261 and ACK according to section 17.1.1.3
-        #            self.parameters["cseq"] += 1
-        #            m["CSeq"] = "{} {}".format(self.parameters["cseq"], m.method)
         self.link.send(m.contents())
         return m
 
@@ -444,6 +473,7 @@ class SipEndpoint(object):
                                                           message_type,
                                                           transaction["method"])
             self.update_to_tag(inmessage.get_dialog())
+            self.free_resources(inmessage)
         self.set_dialog(inmessage.get_dialog())
         return inmessage
 
@@ -521,3 +551,21 @@ class SipEndpoint(object):
             self.re_register_timer.cancel()
         flow.unregister(self)
         self.registered = False
+
+    def free_resources(self, message):
+        """ Free memory allocated to a cleared call """
+
+        status_or_method = message.get_status_or_method()
+        cseq_method = message["CSeq"].split()[1]
+        call_ended_successfully = status_or_method.startswith("2") and cseq_method in ("BYE", "CANCEL")
+        call_rejected_with_error = status_or_method[0] in "3456"
+        if call_ended_successfully or call_rejected_with_error:
+            dialog = message.get_dialog()
+            with self.lock:
+                dialog_index = self.dialogs.index(dialog)
+                # self.dialogs.pop(dialog_index)
+                # self.requests.pop(dialog_index)
+                # self.tags.pop(dialog_hash(dialog))
+                # self.last_messages_per_dialog.pop(dialog_index)
+                print("Cleared call number", dialog_index, "after message", status_or_method, "with cseq", cseq_method,
+                      ". Sizes are", len(self.dialogs), len(self.requests), len(self.tags))
